(*---------------------------------------------------------------------------*
  $Change$
  Copyright (c) 2006-2010, James H. Woodyatt
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** Core I/O event driver framework.  *)

(** {6 Overview}
    
    The core types, functions and classes for composing I/O event driven
    application services using the [Iom_gadget] monad are defined.
    
    Gadgets derived from [Iom_stream] process sequences of typed events with
    flow control signaling.  Simplex streams are either {i consumers}, i.e.
    they receive an event sequence while they are ready, or they are
    {i producers}, i.e. they send an event sequence unless they are blocked.
    Duplex streams both send and receive event sequences.
    
    All stream gadgets receive control events and send notification events.
    Control and notification events are polymorphic variant types.  By
    convention, stream gadgets do not process control events they don't
    recognize.  Stream gadgets respond to the [`Stop] control event by
    immediately (perhaps abortively) releasing all contained resources and
    terminating its processing.  Stream gadgets send [`Failed of exn]
    notification events to indicate they have unexpectedly terminated all
    processing.  Flow control events, i.e. [`Ready] and [`Wait] are sent and/or
    received as either control or notification events, as appropriate.
    
    Stream gadget may process events of any type, but [class event] is provided
    as a convenient base class for streams of events that may terminate in an
    orderly way, e.g. octet streams connected to pipes, files or TCP sockets.
*)

(** {6 Event Types} *)

(** A control event that stops a gadget, abortively if necessary. *)
type stop = [ `Stop ]

(** A notification event that indicates unexpected termination. *)
type failed = [ `Failed of exn ]

(** A control or notification event signaling readiness to process further
    events in the stream.
*)
type ready = [ `Ready ]

(** A control or notification event signaling that further events in the stream
    will be queued for later processing.
*)
type wait = [ `Wait ]

(** A convenient type that sums the flow control events. *)
type readywait = [ ready | wait ]

(** A convenience type to represent the core control events of producer stream
    gadgets.
*)
type flowcontrol = [ stop | readywait ]

(** A convenient type to represent the core notification events of consumer
    stream gadgets.
*)
type flownotify = [ failed | readywait ]

(** A variant type that indicates whether an event is the final event before
    the sequence orderly terminates.
*)
type more = Last | More

(** A convenient class type for objects representing events in sequences that
    may terminate in an orderly way, e.g. octet streams connected to pipes,
    files or TCP sockets.
*)
class type event = object
    
    (** Returns [Last] if this is the final event before orderly termination of
        the event sequence.  Otherwise, returns [More].
    *)
    method more: more
end

(** {6 Plugs and Jacks} *)

(** A pad type for all stream gadgets. *)
type ('x, 'c, 'n) pad = 'x * ('c, 'n) Iom_gadget.pad
    constraint 'c = [> stop ]
    constraint 'n = [> failed ]

(** A fix type for all stream gadgets. *)
type ('x, 'c, 'n) fix = 'x * ('c, 'n) Iom_gadget.fix
    constraint 'c = [> stop ]
    constraint 'n = [> failed ]

(** A specialization of the pad type for simplex producer stream gadgets. *)
type ('i, 'c, 'n) ipad = ('i Iom_gadget.tx, 'c, 'n) pad
    constraint 'c = [> readywait ]
    
(** A specialization of the fix type for simplex producer stream gadgets. *)
type ('i, 'c, 'n) ifix = ('i Iom_gadget.rx, 'c, 'n) fix
    constraint 'c = [> readywait ]

(** A specialization of the pad type for simplex consumer stream gadgets. *)
type ('o, 'c, 'n) opad = ('o Iom_gadget.rx, 'c, 'n) pad
    constraint 'n = [> readywait ]

(** A specialization of the fix type for simplex consumer stream gadgets. *)
type ('o, 'c, 'n) ofix = ('o Iom_gadget.tx, 'c, 'n) fix
    constraint 'n = [> readywait ]

(** A specialization of the pad type for duplex stream gadgets. *)
type ('i, 'o, 'c, 'n) iopad = (('o, 'i) Iom_gadget.pad, 'c, 'n) pad
    constraint 'c = [> readywait ]
    constraint 'n = [> readywait ]

(** A specialization of the fix type for duplex stream gadgets. *)
type ('i, 'o, 'c, 'n) iofix = (('o, 'i) Iom_gadget.fix, 'c, 'n) fix
    constraint 'c = [> readywait ]
    constraint 'n = [> readywait ]

(** Use [ifix f] to create a harness of wires, with an [ipad] at one end and
    a matching [ifix] at the other.  The pad is applied to [f] to obtain a
    gadget object, its start method is applied, and the [ifix] is returned.
*)
val ifix:
  (('i, 'c, 'n) ipad -> #Iom_gadget.start) -> ('i, 'c, 'n) ifix Iom_gadget.t

(** Use [ofix f] to create a harness of wires, with an [opad] at one end and
    a matching [ofix] at the other.  The pad is applied to [f] to obtain a
    gadget object, its start method is applied, and the [ofix] is returned.
*)
val ofix:
  (('o, 'c, 'n) opad -> #Iom_gadget.start) -> ('o, 'c, 'n) ofix Iom_gadget.t

(** Use [iofix f] to create a harness of wires, with an [iopad] at one end
    and a matching [iofix] at the other.  The pad is applied to [f] to obtain
    a gadget object, its start method is applied, and the [iofix] is returned.
*)
val iofix:
  (('i, 'o, 'c, 'n) iopad -> #Iom_gadget.start) ->
  ('i, 'o, 'c, 'n) iofix Iom_gadget.t


(** {6 Gadgets }
    
    Commonly used gadget classes that operate on the standard stream control
    and notify events.  The classes below are intended to be used as the basis
    for a stack of gadgets, with a driver at the bottom, zero or more processes
    above the driver, and an engine at the top.
*)

(** The virtual base class of all stream gadget states. *)
class virtual ['c, 'n] substate: object('self)
    constraint 'self = #Iom_gadget.next
    constraint 'c = [> stop ]
    constraint 'n = [> failed ]
    
    method private fail: 'a. exn -> 'a Iom_gadget.t
    method private stop: 'a. 'a Iom_gadget.t
end

(** Use [inherit control pad] to derive a stream gadget state that receives
    events on the control wire in [pad].
*)
class virtual ['c, 'n] control:
    ('c #Iom_gadget.rx * 'n #Iom_gadget.tx) ->
    object
        inherit ['c, 'n] substate
        
        (** When events are received on the control wire in [pad], they are
            dispatched to [self#control event] and the returned action is
            evaluated.
        *)
        method private control: 'c -> unit Iom_gadget.t
        
        (** Invoke the [guard] method on the inherited base class to guard on
            the control events for [pad].
        *)
        method private guard: unit Iom_gadget.guard
    end

(** Use [inherit notify fix] to derive a stream gadget state that receives
    events on the notification wire in [fix].
*)
class virtual ['c, 'n] notify:
    ('n #Iom_gadget.rx * 'c #Iom_gadget.tx) ->
    object
        inherit ['c, 'n] substate
        
        (** When events are received on the notification wire in [fix], they
            are dispatched to [self#notify event] and the returned action is
            evaluated.
        *)
        method private notify: 'n -> unit Iom_gadget.t
        
        (** Invoke the [guard] method on the inherited base class to guard on
            the notification events for [fix].
        *)
        method private guard: unit Iom_gadget.guard
    end

(** Use [inherit engine fix] to derive a stream gadget that receives events
    on the notification wire and sends events on the control wire in [fix].
*)
class ['c, 'n] engine:
    ('n #Iom_gadget.rx * 'c #Iom_gadget.tx) ->
    object
        inherit ['c, 'n] notify
        inherit Iom_gadget.start
        inherit Iom_gadget.next
    end

(** Use [inherit driver pad] to derive a stream gadget that receives events
    on the control wire and sends events on the notification wire in [pad].
*)
class ['c, 'n] driver:
    ('c #Iom_gadget.rx * 'n #Iom_gadget.tx) ->
    object
        inherit ['c, 'n] control
        inherit Iom_gadget.start
        inherit Iom_gadget.next
    end

(** Use [inherit process fix pad] to derive a stream gadget that processes
    events received on the control wire in [pad] and forwards them on the
    control wire in [fix].  Likewise, it processes events received on the
    notification wire in [fix] and forwards them on the notification wire in
    [pad].
*)
class ['c, 'n] process:
    ('n #Iom_gadget.rx * 'c #Iom_gadget.tx) ->
    ('c #Iom_gadget.rx * 'n #Iom_gadget.tx) ->
    object
        inherit ['c, 'n] control
        inherit ['c, 'n] notify
        inherit Iom_gadget.start
        inherit Iom_gadget.next
    end

(*--- $File$ ---*)
